//@ts-check
"use strict";

const tileSize = 32;
const inverseTileSize = 1 / tileSize;
const chunkSize = 32;

/**
@typedef AmmoSlot
@type {{
  ammoDesc: AmmoDesc;
  quantity: number;
}}
*/

/**
@typedef ItemSlot
@type {{
  item?: ItemName,
}}
 */

/**
@typedef Stats
@type {{
  health: number;
  maxHealth: number;
  armor: number;
  maxArmor: number;

  ammoSlots: AmmoSlot[];
  currentAmmo: number;

  ingredientSlots: IngredientSlot[];
  recipeSlots: RecipeDesc[];
  itemSlots: ItemSlot[];

  // unique ID of hovered item
  hovered: Interactive | null;

  // Quest stuff
  baddiesKilled: { [K in BaddieName]?: boolean; };
  killedAllBaddiesOnce: boolean;
  shipRepaired: boolean;
  sheepKilled: number;
  questsCompleted: { [K in QuestName]?: boolean };
  questsInProgress: { [K in QuestName]?: boolean };
  savedNPCs: { [K in NPCName]?: boolean },
  followers: { [K in NPCName]?: boolean };
}}
*/

/**
 * @typedef QuestState
 * @type {{
 *   grandmaGotMinion: boolean,
 *   sheepKilled: number,
 *   shipRepaired: boolean,
 * }}
 */

/**
 * @typedef IngredientSlot
 * @type {{
 *  plantDesc: PlantDesc,
 *  quantity: number,
 * }}
 */

/**
@typedef InGame
@type {{
  pixiRoot: PIXI.Container,
  pixiWorldContainer: PIXI.Container;
  camera: Camera;
  weather: Weather;
  fade: Fade;
  target: Target;
  stats: Stats;
  hud: Hud;

  gameOver: boolean;

  input: {
    mouse: { x: number, y: number, buttons: boolean[], presses: boolean[]; releases: boolean[] },
    actions: { moveUp: boolean; moveLeft: boolean; moveDown: boolean; moveRight: boolean; }
  };

  tileset: TileSet;
  worlds: { [worldName: string]: World };
  activeWorld: World;
  baxton: Baxton;
  dialog: DialogState,
}}
*/

/** @type InGame */
let inGame = null;

const characterNames = [""];
const worldNames = ["overworld", "underworld"];

const baddies = {};
const friendlies = {};

const mapsFolderName = !location.search.includes("dev") ? "maps" : "devMaps";

function registerInGameResources() {
  let resourcePaths = [
    "inGame/target.png",
    "inGame/fade.png",
    "inGame/weathers/fog.png",
    "inGame/shadow.png",

    // Baxton
    "inGame/baxton/baxtonBody.png",
    "inGame/baxton/baxtonBody.sprite.json",
    "inGame/baxton/baxtonArmIdle.png",
    "inGame/baxton/baxtonArmFiring.png",

    // Fxs
    "inGame/fxs/vanish.png",
    "inGame/fxs/vanish.sprite.json",
    "inGame/fxs/run.png",
    "inGame/fxs/run.sprite.json",
    "inGame/fxs/heartPickUp.png",
    "inGame/fxs/armorPickUp.png",

    // Special
    "inGame/boat/boat.png",
    "inGame/boat/boat.sprite.json",

    // HUD
    "inGame/hud/dialog.png",
    "inGame/hud/health.png",
    "inGame/hud/bagBg.png",
    "inGame/hud/ammoSlot.png",
    "inGame/hud/ammo/pea.png",
    "inGame/hud/ammo/goo.png",
    "inGame/hud/ammo/onion.png",
    "inGame/hud/ammo/leek.png",
    "inGame/hud/ammo/salad.png",
    "inGame/hud/recipeSlot.png",
    "inGame/hud/craftPointer.png",
    "inGame/hud/craftWheel.png",
    "inGame/hud/ingredients/leaf.png",
    "inGame/hud/ingredients/flower.png",
    "inGame/hud/ingredients/dragonfly.png",
    "inGame/hud/ingredients/eyes.png",
    "inGame/hud/ingredients/nose.png",
  ];
  for (const k of Object.keys(baddieDescs)) {
    const baddieDesc = baddieDescs[k];
    if (baddieDesc.textureVariants == null) {
      resourcePaths.push(`inGame/baddies/${k}/body.png`);
    } else {
      for (let textureIndex = 1; textureIndex <= baddieDesc.textureVariants; textureIndex++) {
        resourcePaths.push(`inGame/baddies/${k}/body${textureIndex}.png`);
      }
    }
    resourcePaths.push(`inGame/baddies/${k}/body.sprite.json`);
  }
  for (const k of Object.keys(ammoDescs)) {
    let spriteName = ammoDescs[k].sprite;
    if (spriteName == null) spriteName = k;

    // NOTE: A couple bullets use the same sprites
    if (resourcePaths.indexOf(`inGame/bullets/${spriteName}.png`) !== -1) continue;

    resourcePaths.push(`inGame/bullets/${spriteName}.png`);
    resourcePaths.push(`inGame/bullets/${spriteName}.sprite.json`);
    resourcePaths.push(`inGame/bullets/${spriteName}Blast.png`);
    resourcePaths.push(`inGame/bullets/${spriteName}Blast.sprite.json`);
    resourcePaths.push(`inGame/bullets/${spriteName}Hit.png`);
    resourcePaths.push(`inGame/bullets/${spriteName}Hit.sprite.json`);
  }
  for (const k of Object.keys(itemDescs)) {
    resourcePaths.push(`inGame/items/${k}.png`);
  }
  for (const k of Object.keys(npcDescs)) {
    resourcePaths.push(`inGame/npcs/${k}/body.png`);
    resourcePaths.push(`inGame/npcs/${k}/body.sprite.json`);
  }
  for (const k of Object.keys(plantDescs)) {
    resourcePaths.push(`inGame/plants/${k}.png`);
    resourcePaths.push(`inGame/plants/${k}.sprite.json`);
  }

  app.loader.add(resourcePaths);

  // Maps
  app.loader.add([`inGame/${mapsFolderName}/tileset.tsx`, `inGame/${mapsFolderName}/tileset.png`]);
  app.loader.add(worldNames.map(worldName => `inGame/${mapsFolderName}/${worldName}.tmx`));
}

function enterInGame() {
  game.cursor.sprite.visible = false;

  const pixiRoot = new PIXI.Container();
  game.screenPixiRoot.addChild(pixiRoot);

  const pixiWorldContainer = new PIXI.Container();
  pixiWorldContainer.scale.set(tileSize);
  pixiRoot.addChild(pixiWorldContainer);

  const baxton = createBaxton();

  /** @type {Stats} */
  const stats = {
    health: 3,
    maxHealth: 3,
    armor: 0,
    maxArmor: 5,
    ammoSlots: [
      { ammoDesc: ammoDescs.pea, quantity: 20 },
      { ammoDesc: ammoDescs.goo, quantity: 5 },
      null,
    ],
    currentAmmo: 0,
    ingredientSlots: [
      {
        plantDesc: plantDescs.leaf,
        quantity: 0,
      },
      {
        plantDesc: plantDescs.flower,
        quantity: 0,
      },
      {
        plantDesc: plantDescs.dragonfly,
        quantity: 0,
      },
      {
        plantDesc: plantDescs.eyes,
        quantity: 0,
      },
      {
        plantDesc: plantDescs.nose,
        quantity: 0,
      },
    ],
    recipeSlots: [
      {
        ingredients: [
          plantDescs.leaf,
        ],
        result: ammoDescs.pea,
        amount: 10,
      },
      {
        ingredients: [
          plantDescs.leaf,
          plantDescs.flower,
        ],
        result: ammoDescs.goo,
        amount: 6,
      },
      {
        ingredients: [
          plantDescs.flower,
          plantDescs.dragonfly,
        ],
        result: ammoDescs.salad,
        amount: 1,
      },
      {
        ingredients: [
          plantDescs.dragonfly,
          plantDescs.eyes,
          plantDescs.nose,
        ],
        result: ammoDescs.leek,
        amount: 1,
      },
      {
        ingredients: [
          plantDescs.flower,
          plantDescs.eyes,
          plantDescs.nose,
        ],
        result: ammoDescs.onion,
        amount: 1,
      },
    ],
    sheepKilled: 0,
    questsCompleted: {},
    questsInProgress: {},
    itemSlots: [
      {}, {}, {}, {},
      {}, {}, {}, {},
    ],
    baddiesKilled: {},
    killedAllBaddiesOnce: false,
    shipRepaired: false,
    savedNPCs: {},
    followers: {},
    hovered: null,
  };

  inGame = {
    pixiRoot,
    pixiWorldContainer,
    camera: createCamera(),
    weather: null,
    fade: createFade(),
    target: null,

    stats,
    hud: null,

    gameOver: false,

    input: {
      mouse: { x: 0, y: 0, buttons: [], presses: [], releases: [] },
      actions: { moveUp: false, moveLeft: false, moveDown: false, moveRight: false }
    },

    tileset: loadTileset(),
    worlds: {},
    activeWorld: null,
    baxton,
    dialog: null,
  };

  inGame.weather = createWeather();
  inGame.hud = createHud();
  updateHudHealth();
  updateHudAmmo();
  updateHudIngredients();

  inGame.target = createTarget();
  inGame.dialog = createDialog();

  for (const worldName of worldNames) inGame.worlds[worldName] = loadWorld(worldName);

  let worldName = game.searchParams.get("world") || "overworld";
  let spawnpoint = game.searchParams.get("spawn") || "spawnpoint";
  setActiveWorld(worldName, spawnpoint);

  document.addEventListener("keydown", onInGameKeyDown);
  document.addEventListener("keyup", onInGameKeyUp);
  document.addEventListener("pointermove", onInGamePointerMove);
  document.addEventListener("pointerdown", onInGamePointerDown);
  document.addEventListener("pointerup", onInGamePointerUp);
  document.addEventListener("wheel", onInGameWheel, {
    passive: false,
  });
  //setupAutoPause();
  app.ticker.add(animateInGame);
}

/** @param {() => void} enter - Enter function for next scene */
function leaveInGame(enter) {
  game.cursor.sprite.visible = true;

  setActiveWorld(null, null);

  inGame.baxton.bodySprite.destroy();
  inGame.baxton.armSprite.destroy();
  inGame.baxton.shadowSprite.destroy();

  for (let worldName in inGame.worlds) {
    const world = inGame.worlds[worldName];

    for (let layerName in world.pixiLayers) {
      const tileLayersContainer = world.pixiLayers[layerName].tileLayersContainer;
      for (let tileLayer of tileLayersContainer.children) tileLayer.destroy();
    }

    for (let baddie of world.baddies) {
      baddie.bodySprite.destroy();
      baddie.shadowSprite.destroy();
    }
  }

  document.removeEventListener("keydown", onInGameKeyDown);
  document.removeEventListener("keyup", onInGameKeyUp);
  document.removeEventListener("pointermove", onInGamePointerMove);
  document.removeEventListener("pointerdown", onInGamePointerDown);
  document.removeEventListener("pointerup", onInGamePointerUp);
  document.removeEventListener("wheel", onInGameWheel);
  //teardownAutoPause();
  app.ticker.remove(animateInGame);

  clearFade();
  inGame.pixiRoot.destroy();
  inGame = null;

  enter();
}

/** @param {KeyboardEvent} event */
function onInGameKeyDown(event) {
  const { input, hud } = inGame;

  // Note: `event.code` are *physical key codes*, they don't
  // depend on keyboard layouts, so no need to deal with ZQSD
  switch (event.code) {
    // case "Escape": leaveInGame(enterMainMenu); return;
    case "KeyW": input.actions.moveUp = true; break;
    case "KeyA": input.actions.moveLeft = true; break;
    case "KeyS": input.actions.moveDown = true; break;
    case "KeyD": input.actions.moveRight = true; break;
    default: console.debug(`keydown ${event.code}`);
  }

  if (inGame.dialog.currentDialog) {
    switch (event.code) {
      case "Digit1": dispatchPlayerAction("dialogOption1"); break;
      case "Digit2": dispatchPlayerAction("dialogOption2"); break;
      case "Digit3": dispatchPlayerAction("dialogOption3"); break;
      case "Digit4": dispatchPlayerAction("dialogOption4"); break;
      case "Digit5": dispatchPlayerAction("dialogOption5"); break;
      case "Space": case "KeyE": dispatchPlayerAction("dialogNext"); break;
      default: break;
    }
  } else {
    switch (event.code) {
      case "Digit1": dispatchPlayerAction("selectAmmo1"); break;
      case "Digit2": dispatchPlayerAction("selectAmmo2"); break;
      case "Digit3": dispatchPlayerAction("selectAmmo3"); break;
      case "Space": case "KeyE": dispatchPlayerAction("use"); break;
      case "ShiftLeft": dispatchPlayerAction("craftStart"); break;
      default: break;
    }
  }
}

function setInGameOver() {
  if (inGame.gameOver) return;

  inGame.gameOver = true;
  inGame.hud.isCrafting = false;

  startFadeOut();
}

/** @param {KeyboardEvent} event */
function onInGameKeyUp(event) {
  const { input, hud } = inGame;

  switch (event.code) {
    case "KeyW": input.actions.moveUp = false; break;
    case "KeyA": input.actions.moveLeft = false; break;
    case "KeyS": input.actions.moveDown = false; break;
    case "KeyD": input.actions.moveRight = false; break;
    case "ShiftLeft": dispatchPlayerAction("craftStop"); break;
  }
}

/** @param {MouseEvent} event */
function onInGamePointerMove(event) {
  inGame.input.mouse.x = event.clientX;
  inGame.input.mouse.y = event.clientY;
}

/** @param {MouseEvent} event */
function onInGamePointerDown(event) {
  inGame.input.mouse.buttons[event.button] = true;
  inGame.input.mouse.presses[event.button] = true;
}

/** @param {MouseEvent} event */
function onInGamePointerUp(event) {
  inGame.input.mouse.buttons[event.button] = false;
  inGame.input.mouse.releases[event.button] = true;
}

/** @param {WheelEvent} event */
function onInGameWheel(event) {
  // Never scroll
  event.preventDefault();

  // TODO: ui action next/prev

  if (event.deltaY < 0) {
    dispatchPlayerAction("prevAmmo");
  } else {
    dispatchPlayerAction("nextAmmo");
  }
}

function animateInGame() {
  const dt = app.ticker.deltaMS / 1000;

  updateNearbyChunks(dt);

  animateWorldTiles(dt);
  updateFxs(dt);
  updateWeather(dt);

  updateTarget(dt);
  updateInteractives(dt);

  updateBaddieSpawners(dt);
  updateBaddies(dt);
  updateNPCs(dt);
  updateBullets(dt);
  updateItems(dt);

  updateDialog(dt);
  updateBaxton(dt);
  updateCamera(dt);
  updateBoat(app.ticker.deltaTime);

  animateWorldTitle(dt);
  animateFade(dt);

  updateHud(dt);

  inGame.input.mouse.presses.length = 0;
  inGame.input.mouse.releases.length = 0;

  if (inGame.gameOver && !isFading()) {
    leaveInGame(enterGameOver);
    return;
  }
}

function updateNearbyChunks(dt) {
  const baxtonTilePos = { x: Math.floor(inGame.baxton.pos.x), y: Math.floor(inGame.baxton.pos.y) };

  /** @type {Set<LogicChunk>} */
  const nearbyChunks = new Set();

  for (let i = -1; i <= 1; i++)
    for (let j = -1; j <= 1; j++)
      nearbyChunks.add(inGame.activeWorld.getEntityChunkFor(baxtonTilePos.x + i * chunkSize / 2, baxtonTilePos.y + j * chunkSize / 2));

  for (const chunk of nearbyChunks) {
    for (const teleporter of chunk.teleporters) {
      if (teleporter.rect.contains(inGame.baxton.pos.x, inGame.baxton.pos.y)) {
        setActiveWorld(teleporter.worldName, teleporter.targetName);
      }
    }

    for (const plant of chunk.plants) {
      const dist = distxy(plant.pos, inGame.baxton.pos);

      if (dist <= 2) {
        if (plant.growthCountdown == 0) {
          let slot = inGame.stats.ingredientSlots.find(slot => slot.plantDesc === plant.plantDesc);
          if (slot) {
            getSound(`pickUp.${plant.plantDesc.id}`).play();
            slot.quantity += 1;
            cutPlant(plant);
            updateHudIngredients();
          } else {
            console.warn(`Cannot cut ${plant.plantDesc.id}: no ingredient slot`);
          }
        }
      } else if (plant.growthCountdown > 0) {
        plant.growthCountdown = Math.max(0, plant.growthCountdown - dt);
        if (plant.growthCountdown === 0) {
          growPlant(plant);
        }
      }
    }
  }
}

/**
 * @returns {AmmoDesc} - Ammo that can be shot, or undefined if no ammo left in selected slot
 */
function tryShoot() {
  let slot = inGame.stats.ammoSlots[inGame.stats.currentAmmo];
  if (!slot) {
    // nothing in slot
    return;
  }

  if (slot.quantity <= 0) {
    // out of ammo
    return;
  }

  // Update: we can craft some!
  slot.quantity--;
  updateHudAmmo();

  return slot.ammoDesc;
}

/**
@typedef PlayerAction
@type {
  "selectAmmo1" | "selectAmmo2" | "selectAmmo3" |
  "nextAmmo" | "prevAmmo" | "craftStart" | "craftStop" | "use" |
  "dialogOption1" | "dialogOption2" | "dialogOption3" | "dialogOption4" | "dialogOption5" | "dialogNext"
}
 */

/** @param {PlayerAction} action */
function dispatchPlayerAction(action) {
  const { stats } = inGame;
  if (inGame.gameOver) {
    return;
  }

  switch (action) {
    case "selectAmmo1": stats.currentAmmo = 0; updateHudAmmo(); break;
    case "selectAmmo2": stats.currentAmmo = 1; updateHudAmmo(); break;
    case "selectAmmo3": stats.currentAmmo = 2; updateHudAmmo(); break;
    case "prevAmmo":
      stats.currentAmmo = clamp(stats.currentAmmo - 1, 0, stats.ammoSlots.length - 1);
      updateHudAmmo();
      break;
    case "nextAmmo":
      stats.currentAmmo = clamp(stats.currentAmmo + 1, 0, stats.ammoSlots.length - 1);
      updateHudAmmo();
      break;
    case "craftStart":
      inGame.hud.isCrafting = true;
      break;
    case "craftStop":
      inGame.hud.isCrafting = false;
      break;
    case "use":
      doUse();
      break;
    case "dialogNext":
      selectDialogOption(null);
      break;
    case "dialogOption1":
      selectDialogOption(0);
      break;
    case "dialogOption2":
      selectDialogOption(1);
      break;
    case "dialogOption3":
      selectDialogOption(2);
      break;
    case "dialogOption4":
      selectDialogOption(3);
      break;
    case "dialogOption5":
      selectDialogOption(4);
      break;
    default:
      throw new UnreachableCaseError(action);
  }
}

function doUse() {
  const { stats } = inGame;

  if (stats.hovered) {
    if (stats.hovered.kind === "item") {
      let item = inGame.activeWorld.items.find((item) => item.id === stats.hovered.id);
      if (item) {
        if (pickupItem(item.desc.id)) {
          destroyItem(item);
        }
      }
    } else if (stats.hovered.kind === "npc") {
      let npc = inGame.activeWorld.npcs.find((npc) => npc.id === stats.hovered.id);
      if (!npc) {
        return;
      }

      let dialog = initialDialog(npc);
      if (dialog) {
        setCurrentDialog(dialog);
      }
    }
  }
}
